/*!
 * \file variable_adjoint_tne2.cpp
 * \brief Definition of the solution fields.
 * \author Aerospace Design Laboratory (Stanford University) <http://su2.stanford.edu>.
 * \version 2.0.8
 *
 * Stanford University Unstructured (SU2).
 * Copyright (C) 2012-2013 Aerospace Design Laboratory (ADL).
 *
 * SU2 is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * SU2 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with SU2. If not, see <http://www.gnu.org/licenses/>.
 */

#include "../include/variable_structure.hpp"

CAdjTNE2EulerVariable::CAdjTNE2EulerVariable(void) : CVariable() {
  
  /*--- Array initialization ---*/
	Psi = NULL;
	ForceProj_Vector = NULL;
  
}

CAdjTNE2EulerVariable::CAdjTNE2EulerVariable(double *val_psirho,
                                             double *val_phi,
                                             double val_psie,
                                             double val_psieve,
                                             unsigned short val_ndim,
                                             unsigned short val_nvar,
                                             CConfig *config) : CVariable(val_ndim,val_nvar, config) {

	unsigned short iDim, iMesh, iSpecies, iVar, nMGSmooth = 0;
  
  /*--- Get properties ---*/
  nSpecies = config->GetnSpecies();
  nDim     = val_ndim;
  
  /*--- Array initialization ---*/
	Psi              = NULL;
	ForceProj_Vector = NULL;
  
	/*--- Allocate residual structures ---*/
  Res_TruncError = new double [nVar];
  for (iVar = 0; iVar < nVar; iVar++) {
		Res_TruncError[iVar] = 0.0;
	}
  
  /*--- Only for residual smoothing (multigrid) ---*/
	for (iMesh = 0; iMesh <= config->GetMGLevels(); iMesh++)
		nMGSmooth += config->GetMG_CorrecSmooth(iMesh);
  
  if (nMGSmooth > 0) {
    Residual_Sum = new double [nVar];
    Residual_Old = new double [nVar];
  }
	
	/*--- Allocate undivided laplacian ---*/
	if (config->GetKind_ConvNumScheme_AdjTNE2() == SPACE_CENTERED)
		Undivided_Laplacian = new double [nVar];
  
  /*--- Allocate limiter ---*/
	if (config->GetKind_ConvNumScheme_AdjTNE2() == SPACE_UPWIND) {
		Limiter      = new double [nVar];
		Solution_Max = new double [nVar];
		Solution_Min = new double [nVar];
		for (iVar = 0; iVar < nVar; iVar++) {
			Limiter[iVar]      = 0.0;
			Solution_Max[iVar] = 0.0;
			Solution_Min[iVar] = 0.0;
		}
	}
  
  /*--- Allocate and initialize solution ---*/
  for (iSpecies = 0; iSpecies < nSpecies; iSpecies++) {
    Solution[iSpecies]     = val_psirho[iSpecies];
    Solution_Old[iSpecies] = val_psirho[iSpecies];
  }
  for (iDim = 0; iDim < nDim; iDim++) {
    Solution[nSpecies+iDim]     = val_phi[iDim];
    Solution_Old[nSpecies+iDim] = val_phi[iDim];
  }
  Solution[nSpecies+nDim]       = val_psie;
  Solution_Old[nSpecies+nDim]   = val_psie;
  Solution[nSpecies+nDim+1]     = val_psieve;
  Solution_Old[nSpecies+nDim+1] = val_psieve;
  
  /*--- Allocate auxiliar vector for sensitivity computation ---*/
	Grad_AuxVar = new double [nDim];
	
	/*--- Allocate and initialize projection vector for wall boundary condition ---*/
	ForceProj_Vector = new double [nDim];
	for (iDim = 0; iDim < nDim; iDim++)
		ForceProj_Vector[iDim] = 0.0;
	
}

CAdjTNE2EulerVariable::CAdjTNE2EulerVariable(double *val_solution,
                                             unsigned short val_ndim,
                                             unsigned short val_nvar,
                                             CConfig *config) : CVariable(val_ndim, val_nvar, config) {
	unsigned short iVar, iDim, iMesh, nMGSmooth = 0;
  bool dual_time = ((config->GetUnsteady_Simulation() == DT_STEPPING_1ST) ||
                    (config->GetUnsteady_Simulation() == DT_STEPPING_2ND));
  
  /*--- Array initialization ---*/
	Psi              = NULL;
	ForceProj_Vector = NULL;
  
	/*--- Allocate residual structures ---*/
  Res_TruncError = new double [nVar];
  
  for (iVar = 0; iVar < nVar; iVar++) {
		Res_TruncError[iVar] = 0.0;
	}
  
  /*--- Only for residual smoothing (multigrid) ---*/
	for (iMesh = 0; iMesh <= config->GetMGLevels(); iMesh++)
		nMGSmooth += config->GetMG_CorrecSmooth(iMesh);
  
  if (nMGSmooth > 0) {
    Residual_Sum = new double [nVar];
    Residual_Old = new double [nVar];
  }
  
	/*--- Allocate undivided laplacian (centered) and limiter (upwind)---*/
	if (config->GetKind_ConvNumScheme_AdjTNE2() == SPACE_CENTERED)
		Undivided_Laplacian = new double [nVar];
	if (config->GetKind_ConvNumScheme_AdjTNE2() == SPACE_UPWIND) {
		Limiter      = new double [nVar];
		Solution_Max = new double [nVar];
		Solution_Min = new double [nVar];
		for (iVar = 0; iVar < nVar; iVar++) {
			Limiter[iVar]      = 0.0;
			Solution_Max[iVar] = 0.0;
			Solution_Min[iVar] = 0.0;
		}
	}
  
	/*--- Solution initialization ---*/
	for (iVar = 0; iVar < nVar; iVar++) {
		Solution[iVar] = val_solution[iVar];
		Solution_Old[iVar] = val_solution[iVar];
	}
  
	/*--- Allocate and initializate solution for dual time strategy ---*/
	if (dual_time) {
		Solution_time_n = new double [nVar];
		Solution_time_n1 = new double [nVar];
    
		for (iVar = 0; iVar < nVar; iVar++) {
			Solution_time_n[iVar] = val_solution[iVar];
			Solution_time_n1[iVar] = val_solution[iVar];
		}
	}
  
  /*--- Allocate auxiliar vector for sensitivity computation ---*/
	Grad_AuxVar = new double [nDim];
	
	/*--- Allocate and initializate projection vector for wall boundary condition ---*/
	ForceProj_Vector = new double [nDim];
	for (iDim = 0; iDim < nDim; iDim++)
		ForceProj_Vector[iDim] = 0.0;
  
}

CAdjTNE2EulerVariable::~CAdjTNE2EulerVariable(void) {
	
	if (Psi               != NULL) delete [] Psi;
	if (ForceProj_Vector  != NULL) delete [] ForceProj_Vector;
  
}

void CAdjTNE2EulerVariable::SetPrimVar_Compressible(double val_adjlimit) {
	unsigned short iVar;
  bool check_dens = false;
  
  check_dens = (fabs(Solution[0]) > val_adjlimit);             // Check the adjoint density
  
  /*--- Check that the solution has a physical meaning ---*/
  if (check_dens) {
    
    /*--- Copy the old solution ---*/
    for (iVar = 0; iVar < nVar; iVar++)
      Solution[iVar] = Solution_Old[iVar];
    
  }
  
}


CAdjTNE2NSVariable::CAdjTNE2NSVariable(void) : CAdjTNE2EulerVariable() { }

CAdjTNE2NSVariable::CAdjTNE2NSVariable(double *val_solution,
                                       unsigned short val_ndim,
                                       unsigned short val_nvar,
                                       CConfig *config) : CAdjTNE2EulerVariable(val_solution,
                                                                                val_ndim,
                                                                                val_nvar,
                                                                                config) {
  
}

CAdjTNE2NSVariable::CAdjTNE2NSVariable(double *val_psirho, double *val_phi,
                                       double val_psie, double val_psieve, unsigned short val_ndim,
                                       unsigned short val_nvar, CConfig *config) : CAdjTNE2EulerVariable(val_psirho,
                                                                                                         val_phi,
                                                                                                         val_psie,
                                                                                                         val_psieve,
                                                                                                         val_ndim,
                                                                                                         val_nvar,
                                                                                                         config) {

}

CAdjTNE2NSVariable::~CAdjTNE2NSVariable(void) { }
